/*
 * Copyright (c) 1996-1999 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

/************************************************************************
 * locktab.h: declaration of LockTable class
 * 
 * A LockTable is the reader-writer locking implementation for the
 * thread library's thread table.  It relies on the pthreads 
 * locking primitives.  If USE_PTHREADS_BASED_LIBTHREAD is not true, all 
 * class methods are no-ops.  Some ideas for this code come from Butenhof's
 * _Programming With POSIX Threads_.  (AWL, 1997) --willb, 9/2001
 *
 * NB!  There are no cleanup handlers in this code.  If you use thread
 * cancellation, you deserve what you get.
 *
 * $Revision: 1.1 $
************************************************************************/
#ifndef _thread_src_locktab_h_
#define _thread_src_locktab_h_

#include<stdlib.h>

#define USE_PTHREADS_BASED_LIBTHREAD 1

#if USE_PTHREADS_BASED_LIBTHREAD
#include <pthread.h>
#endif

#define LOCKTAB_FAVOR_READERS 0xda7aba5e
#define LOCKTAB_FAVOR_WRITERS 0xac71014 
#define DEFAULT_SIZE 512

class LockTable {  
 private:
#if USE_PTHREADS_BASED_LIBTHREAD
  int table_size;
  int preference;

  pthread_mutex_t *mutex_list;
  /* we use two condition variables instead of one variable with two
     predicates to avoid spurious wakeups */
  pthread_cond_t *read_cond;
  pthread_cond_t *write_cond;
  int *active_readers; 
  int *active_writers;
  int *waiting_readers;
  int *waiting_writers;

  inline void alloc_storage() {
    mutex_list = (pthread_mutex_t*)
      malloc(sizeof(pthread_mutex_t) * table_size);
    read_cond = (pthread_cond_t*)
      malloc(sizeof(pthread_cond_t) * table_size);
    write_cond = (pthread_cond_t*)
      malloc(sizeof(pthread_cond_t) * table_size);
    active_readers = (int*)malloc(sizeof(int) * table_size);
    active_writers = (int*)malloc(sizeof(int) * table_size);
    waiting_readers = (int*)malloc(sizeof(int) * table_size);
    waiting_writers = (int*)malloc(sizeof(int) * table_size);
  }

  inline void init_structures() {
    register int i = 0;
    for(i = 0; i < table_size; i++) {
      pthread_mutex_init(&mutex_list[i], NULL);
      pthread_cond_init(&read_cond[i], NULL);
      pthread_cond_init(&write_cond[i], NULL);
      active_readers[i] = 0;
      active_writers[i] = 0;
      waiting_readers[i] = 0;
      waiting_writers[i] = 0;
    }
  }
#endif
 public:
  /* The default constructor for LockTable creates a table of the
     default size, favoring writers */

#if USE_PTHREADS_BASED_LIBTHREAD
  LockTable(int size = DEFAULT_SIZE) : table_size(size),
    preference(LOCKTAB_FAVOR_WRITERS) {

    alloc_storage();
    init_structures();
  }

  LockTable(int size, int favor=LOCKTAB_FAVOR_WRITERS) :
    table_size(size), preference(LOCKTAB_FAVOR_WRITERS) {
  
    alloc_storage();
    init_structures();
  }
#else
  LockTable(int size = DEFAULT_SIZE) { }
  LockTable(int size, int favor=LOCKTAB_FAVOR_WRITERS) { }
#endif  
  int start_reading(int which);
  int start_writing(int which);
  
  int stop_reading(int which);
  int stop_writing(int which);
};


#endif // !defined(_thread_src_locktab_h_)
